#include "regexp.h"

using namespace std;

static void setloop(wstring &s, const wchar_t ope);
static wstring cnvset(const wchar_t *s, int &pos, const size_t len);
static void chkset(const wstring s);
static wstring escstr(const wchar_t *s, int &pos, const size_t len);
static void prepare(regexpbuf &e);
vector<pair<int, int> > getbrefpos(regexpbuf &e);
static int fwdacct(regexpbuf &e);
static int revacct(regexpbuf &e);

/* ɽΥѥ 
 * ȹ票󥸥VMɤߤȤѴ
 */

regexpbuf re_compile(const wchar_t *s, const size_t len);
regexpbuf re_compile(const wstring s);

regexpbuf re_compile(const wstring s) {
  return re_compile(s.data(), s.size());
}

regexpbuf re_compile(const wchar_t *s, const size_t len) {
  regexpbuf d;
  int i = 0;
  while (i < len) {
    if (!d.empty()
     && ((s[i] == L'?') || (s[i] == L'*') || (s[i] == L'+'))) {
      setloop(d.back(), s[i]);
      i++;
    } else if (s[i] == L'.') {
      d.push_back(wstring(L"."));
      i++;
    } else if (((i + 1) < len) && (s[i] == '\\') && (s[i + 1] == '(')) {
      d.push_back(wstring(L"("));
      i++;
      i++;
    } else if (((i + 1) < len) && (s[i] == '\\') && (s[i + 1] == ')')) {
      d.push_back(wstring(L")"));
      i++;
      i++;
    } else if (s[i] == L'[') {
      d.push_back(wstring(L"C") + cnvset(s, i, len));
    } else {
      d.push_back(wstring(L"L") + escstr(s, i, len));
    }
  }
  prepare(d);
  return d;
}

/* closureѴ
 * ((c*)*) -> (c*) 
 * ((c*)+) -> (c*)
 * ((c?)?) -> (c?)
 * ((c+)+) -> (c+)
 */

static void setloop(wstring &s, const wchar_t ope) {
  switch (s[0]) {
  case L'*': s[0] = L'*'; break;
  case L'+': s[0] = ((ope == '*') ? L'*' : L'+'); break;
  case L'?': s[0] = ((ope == '?') ? L'?' : L'*'); break;
  default: s = wstring(&ope, 1) + s; break;
  }
}

/* Ѵ
 *
 * ʾη֤ȥʾȹ礻Ѵ
 * Хåե󥹤λֹ
 */

static void prepare(regexpbuf &s) {
  regexpbuf d;
  vector<wchar_t> st;
  int n = 0;
  for (regexpbuf::iterator it = s.begin(); it != s.end(); it++) {
    if ((*it)[0] == L'(') {
      if (++n > 9) {
        throw runtime_error("regexp syntax error:back ref too many");
      }
      wchar_t r = (wchar_t)(L'0' + n);
      st.push_back(r); 
      d.push_back(*it + r);
    } else if ((*it)[0] == L')') {
      if (st.empty()) {
        throw runtime_error("regexp syntax error:back ref unmatch");
      }
      wchar_t r = st.back();
      st.pop_back();
      d.push_back(*it + r);
    } else if ((*it)[0] == L'+') {
      d.push_back(it->substr(1));
      d.push_back(L"*" + it->substr(1));
    } else {
      d.push_back(*it);
    }
  }
  if (!st.empty()) {
    throw runtime_error("regexp syntax error:back ref unmatch");
  }
  s = d;
}

/* Хåե󥹤ʬΥФ */

vector<pair<int, int> > getbrefpos(regexpbuf &e) {
  regexpbuf d;
  vector<pair<int, int> > v;
  v.push_back(pair<int, int>(0, 0));
  int n = 0;
  for (int i = 0; i < e.size(); i++) {
    if (e[i][0] == L'(') {
      v.push_back(pair<int, int>(n - 1, n - 1));
    } else if (e[i][0] == L')') {
      v[e[i][1] - L'0'].second = n - 1;
    } else {
      d.push_back(e[i]);
      n++;
    }
  }
  e = d;
  return v;
}

/* ƬظƽǾȹ礷
 * ĤǽȤʤ뤫Ƚ
 * OR롼ײ黻Ҥ̵Τ
 * nullableǤʤñ㥵
 * ĤʤжȾȹ礷ѥ
 */

static int fwdacct(regexpbuf &s) {
  int n = s.size();
  while (n--) {
    switch (s[n][0]) {
    case L'*':
    case L'?':
      continue;
    case L'+':
    case L'C':
    case L'L':
    case L'.':
      return n;
    }
  }
  return 0; // nullable
}

/* ƬظƵǾȹ礷
 * ĤǽȤʤ뤫Ƚ
 * OR롼ײ黻Ҥ̵Τ
 * nullableǤʤƬñ㥵
 * ĤʤжȾȹ礷ѥ
 */

static int revacct(regexpbuf &s) {
  int n = s.size();
  for (int i = 0; i < n; i++) {
    switch (s[n][0]) {
    case L'*':
    case L'?':
      continue;
    case L'+':
    case L'C':
    case L'L':
    case L'.':
      return n;
    }
  }
  return 0; // nullable
}

/* 饯åȡʸ饹 */

static wstring cnvset(const wchar_t *s, int &pos, const size_t len) {
  wstring d;
  pos++; // skip "[" operator.
  if (pos >= len) {
    throw runtime_error("regexp syntax error:[] unmatched");
  }
  if (s[pos] == L'^') {
    d += L"^";
    pos++;
  }
  int top = pos;
  while (pos < len) {
    if ((pos != top) && (s[pos] == L']')) {
      pos++; // skip "]" operator.
      chkset(d);
      return d;
    } else {
      d += escstr(s, pos, len);
    }
  }
  throw runtime_error("regexp syntax error:[] unmatched");
}

/* 饯åȡʸ饹ʸˡå */

static void chkset(const wstring e) {
  wstring s = e;
  if (s[0] == L'^') s = s.substr(1);
  if (s.size() < 1) {
    throw runtime_error("regexp syntax error:[] empty");
  }
  int l = (int)s.size() - 2;
  for (int i = 1; i < l; i++) {
    if (s[i] == L'-') {
      if ((s[i - 1]) > (s[i + 1])) {
        throw runtime_error("regexp syntax error:[] randge error");
      }
    }
  }
  return; // OK
}

/* Хåå\\׽ */

static wstring escstr(const wchar_t *s, int &pos, const size_t len) {
  wstring d;
  if (s[pos] == L'\\') {
    pos++; // skip backslash operator.
    if (pos < len) {
      switch (s[pos]) { // \x, \u: not compleete
      case L't':
	d += L'\t';
        pos++;
        return d;
      default:
        d += s[pos];
        pos++;
        return d;
      }
    } else {
      throw runtime_error("regexp syntax error:trailing backslash");
    }
  } else {
    d += s[pos];
    pos++;
    return d;
  }
}

/* ʹ߾ȹ票󥸥VM */

/* VM */

vmsys::vmsys() {}
vmsys::vmsys(const regexpbuf s) { init(s); }

/* VM */

void vmsys::init(const regexpbuf s) {
  e = s;
  brefpos = getbrefpos(e);
  acct = fwdacct(e);
  reset();
}

/* VMꥻåȤ */
void vmsys::reset() {
  mc = 0;
  vmcts.clear();
  vmct v;
  pair<int, int> t = pair<int, int>(0, 0);
  v.r.resize(brefpos.size() + 1, t);
  vmcts.push_back(v);
  last.clear();
}

/* Ƥ뤫ɤ֤ */

bool vmsys::matched() {
  return !last.empty();
}

/* Ƥз֤ */

list<vector<pair<int, int> > > vmsys::results() {
  return last;
}

/* Ƥз̤֤ */

vector<pair<int, int> > vmsys::result() {
  return last.front();
}

/* ɽüå */

bool vmsys::istail(const vmct &vm) {
  return (vm.ip < (e.size() - 1)) ? false : true;
}
  
/* ɽñ̤ɤ߿ʤ */

bool vmsys::vmnext(vmct &vm) {
  setbrs(vm);
  setbre(vm);
  if (istail(vm)) {
    return false;
  } else {
    vm.ip++;
    return true;
  }
}

/* ȹ票󥸥ΰ쥵¹Ԥ */

bool vmsys::vmstep(wchar_t c) {
  list<vmct> nexts;
  list<vector<pair<int, int> > > result;
  for (list<vmct>::iterator it = vmcts.begin(); it != vmcts.end(); it++) {
    bool isnext = false;
    for (;;) {
      it->acc = (it->ip >= acct);
      wstring cmd = e[it->ip];
      switch (cmd[0]) {
      case L'?':
        if (matchone(cmd.substr(1), c)) {
          if (vmnext(*it)) {
            nexts.push_back(*it); /* yes:next */
          }
        }
        if (vmnext(*it)) { // skip and branch
          continue;
        }
        break;
      case L'*':
        if (matchone(cmd.substr(1), c)) {
          nexts.push_back(*it); /* repeat */
        }
        if (vmnext(*it)) { // skip and branch
          continue;
        }
        break;
      default:
        if (matchone(cmd, c)) {
          if (vmnext(*it)) {
            nexts.push_back(*it);
          }
        } else it->acc = false;
      }
      if (it->acc) {
        it->r[0].second = mc + 1;
        result.push_back(it->r);
      }
      break;
    }
  }
  if (!result.empty()) { /* acceptable */
    last = result;
  }
  if (nexts.empty()) { /* stop */
    return true;
  } else { /* continue */
    mc++;
    vmcts = nexts;
    return false;
  }
}

/* Хåե󥹤Ƭξȹ */

void vmsys::setbrs(vmct &vm) {
  for (int i = 1; i < brefpos.size(); i++) {
    if (brefpos[i].first == vm.ip) {
      vm.r[i].first = mc + 1;
    }
  }
}

/* Хåե󥹤ξȹ */

void vmsys::setbre(vmct &vm) {
  for (int i = 1; i < brefpos.size(); i++) {
    if (brefpos[i].second == vm.ip) {
      vm.r[i].second = mc + 0;
    }
  }
}

/* ʸåȡʸ饹ξȹ */

bool vmsys::matchset(wstring e, wchar_t c) {
  bool v = false;
  if (e[0] == L'^') {
    v = true;
    e = e.substr(1);
  }
  int l = (int)e.size();
  for (int i = 0; i < l; i++) {
    if ((0 < i) && (i < (l - 1)) && (e[i] == L'-')) {
      if ((e[i - 1] <= c) && (c <= e[i + 1])) {
        return true ^ v;
      }
    } else {
      if (e[i] == c) {
        return true ^ v;
      }
    }
  }
  return false ^ v;
}

/* ʸξȹ */

bool vmsys::matchone(wstring e, wchar_t c) {
  switch(e[0]) {
  case '.': return true;
  case 'L': return e[1] == c;
  case 'C': return matchset(e.substr(1), c);
  default: throw runtime_error("regexp internal error");
  }
}

regexp::regexp() {}
regexp::regexp(const wstring s) { init(s); }

void regexp::init(const wstring s) {
  regexpbuf b = re_compile(s);
  vm.init(b);
}

bool regexp::vmdrive(const wstring s) {
  vm.reset();
  for (wstring::const_iterator it = s.begin(); it != s.end(); it++) {
    if (vm.vmstep(*it)) break;
  }
  return vm.matched();
}

bool regexp::match(const wstring s) {
  if (vmdrive(s)) {
    vector<pair<int, int> > r = vm.result();
    if (r[0].second != s.size()) return false;
  } else return false;
}

bool regexp::search(const wstring s) {
  wstring t = s;
  while (t.size()) {
    if (vmdrive(t)) {
      return true;
    }
    t = t.substr(1);
  }
  return false;
}

vector<std::pair<int, int> > regexp::result() {
  return vm.result();
}

wstring regexp::editref(const wstring s, const wstring d, 
 const vector<pair<int, int> > v) { 
  wstring t;
  int sp, ep;
  int l = (int)d.size();
  for (int i = 0; i < l; i++) {
    if ((i < (l - 1)) && (d[i] == L'\\')) {
      switch (d[i + 1]) {
      case L'&': case L'0':
        sp = v[0].first;
        ep = v[0].second;
        if ((ep - sp) > 0) {
          t += s.substr(sp, ep - sp);
        }
        i++;
        break;
      case L'1': case L'2': case L'3': case L'4': case L'5':
      case L'6': case L'7': case L'8': case L'9':
        sp = v[d[i + 1] - L'0'].first;
        ep = v[d[i + 1] - L'0'].second;
        if ((ep - sp) > 0) {
          t += s.substr(sp, ep - sp);
        }
        i++;
        break;
      case L't':
        t += L'\t';
        i++;
        break;
      case L'n':
        t += L'\n';
        i++;
        break;
      case L'r':
        t += L'\r';
        i++;
        break;
      default:
        t += d[i + 1];
        i++;
        break;
      }
    } else {
      t += d[i];
    }
  }
  return t;
}

wstring regexp::replace(const wstring s, const wstring d) {
  if (search(s)) {
    vector<pair<int, int> > v = vm.result();
    int sp = v[0].first;
    int ep = v[0].second;
    wstring l = ((ep - sp) > 0 ? s.substr(0, sp) : L"");
    wstring r = ((ep < s.size()) ? s.substr(ep) : L"");
    return l + editref(s, d, v) + r;
  } else return s;
}

wstring regexp::replaceall(const wstring s, const wstring d) {
  wstring q, t, l, r;
  for (q = s; search(q); q = r) {
    vector<pair<int, int> > result = vm.result();
    vector<pair<int, int> > v = vm.result();
    int sp = v[0].first;
    int ep = v[0].second;
    l = wstring(q.begin(), &q[sp]);
    r = wstring(&q[ep], q.end());
    t += (l +  editref(q, d, v));
  }
  return t + q;
}

#ifdef SAMPLE // ưǧѥץ

#include <iostream>
#include <fstream>
#include <locale.h>

int main() {
  try {
    setlocale(LC_ALL, "ja_JP.eucJP");
    wstring exp = L"#include[ \t]+<\\([^>]+\\)>";
    cout << "'" << wtom(exp) << "'" << endl;
    regexp re(exp);
    ifstream ifs("regexp.cc");
    while (ifs.good()) {
      string s;
      getline(ifs, s, '\n');
      wstring w = mtow(s);
      if (re.search(w)) {
        vector<pair<int, int> > result = re.result();
        int sp = result[0].first;
        int ep = result[0].second;
        cout << "'" << wtom(w.substr(sp, ep - sp)) << "'";
        if (result.size() > 1) {
          cout << "["
           << result[1].first  << ":"
           << result[1].second << "]";
          string s =
           wtom(wstring(&w[(result[1].first)],
                        &w[(result[1].second)]));
          cout << ":" << "'" << s << "'";
        }
        cout << endl;
        w = re.replace(w, wstring(L"headerfile:\\1(\\0)"));
        cout << wtom(w) << endl;
      }
    }
  } catch (exception &e) {
    cout << e.what() << endl;
  }
}

#endif
